( ASM7r.f, 04sep04cht )
( ARM7 assembler, originally by Chien-ja Wu for win32Forth )
( version 2.02, aligned to GBAEF for validation )
( 	add {{ for LDM and STM )
( version 3.01, updated to subroutine threading )


HEX

VARIABLE h
VARIABLE lasth 0 lasth !                  \ init linkfield address lfa
VARIABLE lastc				\ for RECURSE
VARIABLE hw VARIABLE bi ( for packing )
VARIABLE reg            ( # or registers referred so far )
VARIABLE offset		( -1 for immediate operand )
VARIABLE #offset	( value of immediate operand )
VARIABLE rotate         ( for shifter )

: COMPILE-ONLY 40 lasth @ ram@ XOR lasth @ ram! ;
: IMMEDIATE    80 lasth @ ram@ XOR lasth @ ram! ;

: anew 	0 bi ! 0 reg ! E0000000 hw ! 0 offset ! ;
: org   DUP . CR h ! anew ;
: ALLOT ( n -- ) h +! anew ;

: ,    ( d ) h @ ram! 4 h +! anew ; ( compile instruction )
: #,   ( - ) hw @ , ;		( compile instruction in hw )
: ,w   ( d ) hw @ OR hw ! ;	( add operand to hw )
: ,c   ( d ) 			( change condition field )
	hw @ 0FFFFFFF AND OR hw ! ;
: ,b   ( c ) 			( pack bytes into hw )
	bi @ 0 = IF 1 bi !         hw ! EXIT THEN
	bi @ 1 = IF 2 bi !     100 * ,w EXIT THEN
	bi @ 2 = IF 3 bi !   10000 * ,w EXIT THEN
		    0 bi ! 1000000 * ,w #, ;
: ,b0	bi @ IF #, THEN ;	( zero fill to word boundary )

( define operand stuffing words )

: operand       CONSTANT DOES> R> @ ,w ;
: condition     CONSTANT DOES> R> @ ,c ;
: shifter       CONSTANT DOES> 1F AND 80 * R> @ 20 * + ,w ;
: register      CONSTANT DOES>
        reg @ 0 = IF R> @  1000 * ,w 1 reg +! EXIT THEN ( Rd )
        reg @ 1 = IF R> @ 10000 * ,w 1 reg +! EXIT THEN ( Rn )
        reg @ 2 = IF R> @         ,w 1 reg +! EXIT THEN ( Rm )
        reg @ 3 = IF R> @   100 * ,w 1 reg +! EXIT THEN ( Rs )
        1 R> @ FOR AFT 2* THEN NEXT ,w ( multiple register list )
        ;
: {	( force registers into multiple register list )
        hw @ 10 * F0000 AND
        hw @ FFF00000 AND OR hw !
	4 reg ! ; 
: }	;	( just to pair with { )

( process immediate operands )

: #    ( d ) #offset ! -1 offset ! ;
: #_    # ;

: mem_imm ( -- , with rd,rn )
	offset @ NOT IF EXIT THEN
	#offset @ DUP 
	1000 < NOT ABORT" memory offset out of range "
	,w
	;	
: rotate_imm ( -- , with rd,rn )
	offset @ NOT IF EXIT THEN
        0 rotate !
	#offset @
        0F FOR AFT 
              DUP 100 U<
              IF rotate @ 100 * OR 2000000 OR ,w
                 R> DROP EXIT
              THEN
              DUP 2 LSHIFT SWAP 01E RSHIFT 3 AND +
              1 rotate +! ( continue shifting )
        THEN NEXT
        1 ABORT" immediate out of range"
        ;
: hilo_imm ( -- , for LDRH and STRH )
	offset @ NOT IF EXIT THEN
	#offset @ DUP 
	ABS 100 < NOT ABORT" memory offset out of range "
        DUP 4 LSHIFT F00 AND ,w
	0F AND ,w
        400000 ,w 
        ;

( classes of instructions )

: simple_inst   CONSTANT 
	DOES> R> @ ,w #, ;
: mem_inst   CONSTANT 
	DOES> R> @ ,w 
	reg @ 3 =
	IF 2000000 ,w 
	ELSE mem_imm
	THEN
	#, ;
: hmem_inst   CONSTANT 
	DOES> R> @ ,w 
	hilo_imm
	#, ;
: alu_inst   CONSTANT 
	DOES> R> @ ,w 
	rotate_imm
	#, ;
: tst_inst  CONSTANT 
	DOES> R> @ ,w
	reg @ 2 <
	IF
	    rotate_imm
            hw @ 10 * F0000 AND
            hw @ FFF00FFF AND OR ,
	ELSE
            hw @ 10 * F0000 AND
            hw @ 10 RSHIFT 0F AND OR
            hw @ FFF00FF0 AND OR ,
	THEN
        ;
: mov_inst  CONSTANT 
	DOES> R> @ ,w
	rotate_imm
	reg @ 2 <
	IF  #,
	ELSE 
            hw @ 10 RSHIFT 0F AND
            hw @ FFF0FFF0 AND OR ,
	THEN ;

01A00000 mov_inst ,MOV
01E00000 mov_inst ,MVN
01500000 tst_inst ,CMP
01700000 tst_inst ,CMN
01300000 tst_inst ,TEQ
01100000 tst_inst ,TST

00000000 alu_inst ,AND
00200000 alu_inst ,EOR
00400000 alu_inst ,SUB
00600000 alu_inst ,RSB
00800000 alu_inst ,ADD
00A00000 alu_inst ,ADC
00C00000 alu_inst ,SBC
00E00000 alu_inst ,RSC
01800000 alu_inst ,ORR
01C00000 alu_inst ,BIC

: ,BX   hw @ 0C RSHIFT 0F AND 12FFF10 OR
        hw @ F0000000 AND OR ,
        ;
: ,BLX  hw @ 0C RSHIFT 0F AND 12FFF30 OR
        hw @ F0000000 AND OR ,
        ;
: ,B	( a -- )
	h @ 8 + - 2 RSHIFT FFFFFF AND
	0A000000 OR
        hw @ F0000000 AND OR ,
	;
: ,BL	( a -- )
	h @ 8 + - 2 RSHIFT FFFFFF AND
	0B000000 OR
        hw @ F0000000 AND OR ,
	;

01200070 simple_inst BKPT,
01600010 simple_inst CLZ,
0C000000 simple_inst LDC,
08100000 simple_inst LDM1,
08500000 simple_inst LDM2,
08508000 simple_inst LDM3,
0C000000 simple_inst STC,
08000000 simple_inst STM1,
08400000 simple_inst STM2,

0C100000 simple_inst ,LDC
0C500000 simple_inst ,LDCL
0C000000 simple_inst ,SDC
0C400000 simple_inst ,SDCL

0E000000 simple_inst ,CDP
0E100010 simple_inst ,MRC
0E000010 simple_inst ,MCR

0F000000 simple_inst ,SWI
01000090 simple_inst ,SWP
01400090 simple_inst ,SWPB

: ,MSR  reg @ IF
	     hw @ 0C RSHIFT 0F AND 129F000 OR ,w
	ELSE rotate_imm 321F000 ,w
	THEN
        #,
        ;
010F0000 simple_inst ,MRS

04100000 mem_inst ,LDR
043000F0 mem_inst ,LDRT
04500000 mem_inst ,LDRB
04700000 mem_inst ,LDRBT
04000000 mem_inst ,STR
042000F0 mem_inst ,STRT
04400000 mem_inst ,STRB
04600000 mem_inst ,STRBT
001000B0 hmem_inst ,LDRH
001000D0 hmem_inst ,LDRSB
001000F0 hmem_inst ,LDRSH
000000B0 hmem_inst ,STRH
000000D0 hmem_inst ,STRSB
000000F0 hmem_inst ,STRSH

09900000 simple_inst ,LDMED
09900000 simple_inst ,LDMIB
08900000 simple_inst ,LDMFD
08900000 simple_inst ,LDMIA
09100000 simple_inst ,LDMEA
09100000 simple_inst ,LDMDB
08100000 simple_inst ,LDMFA
08100000 simple_inst ,LDMDA
09800000 simple_inst ,STMFA
09800000 simple_inst ,STMIB
08800000 simple_inst ,STMEA
08800000 simple_inst ,STMIA
09000000 simple_inst ,STMFD
09000000 simple_inst ,STMDB
08000000 simple_inst ,STMED
08000000 simple_inst ,STMDA

( operands )

00000000 operand [-]
00200000 operand [-T]
00400000 operand [-B]
00600000 operand [-BT]
00800000 operand []
00800000 operand [+]
00A00000 operand [+T]
00C00000 operand [+B]
00E00000 operand [+BT]
00A00000 operand [T]
00C00000 operand [B]
00E00000 operand [BT]
01000000 operand [-P]
01200000 operand [-P!]   
01400000 operand [-PB]   
01600000 operand [-PB!]  
01800000 operand [P]   
01A00000 operand [P!]  
01C00000 operand [PB]  
01E00000 operand [PB!] 
01800000 operand [+P]   
01A00000 operand [+PT]
01C00000 operand [+PB]  
01E00000 operand [+PB!]

00400000 operand [^]
00600000 operand [!^]
00400000 operand SPSR,
00000000 operand CPSR,

00000010 operand LSL
00000010 operand ASL
00000030 operand LSR
00000050 operand ASR
00000070 operand ROR
00000060 operand RRX

0 shifter ASL#
0 shifter LSL#
1 shifter LSR#
2 shifter ASR#
3 shifter ROR#

( conditons )

00100000 operand S,
00000000 condition EQ,
10000000 condition NE,
20000000 condition CS,
20000000 condition HS,
30000000 condition CC,
30000000 condition LO,
40000000 condition MI,
50000000 condition PL,
60000000 condition VS,
70000000 condition VC,
80000000 condition HI,
90000000 condition LS,
A0000000 condition GE,
B0000000 condition LT,
C0000000 condition GT,
D0000000 condition LE,
E0000000 condition AL,

( registers )

0 register R0
1 register R1
2 register R2
3 register R3
4 register R4
5 register R5
6 register R6
7 register R7
8 register R8
9 register R9
A register R10
B register R11
C register R12
D register R13
E register R14
F register R15

0 register R0,
1 register R1,
2 register R2,
3 register R3,
4 register R4,
5 register R5,
6 register R6,
7 register R7,
8 register R8,
9 register R9,
A register R10,
B register R11,
C register R12,
D register R13,
E register R14,
F register R15,

( multiply )

: ,NOP ( -- )
        R0, R0  ,MOV ;
: ,MUL   hw @ DUP 10 * F0000 AND
        OVER 100 * F00 AND OR
        OVER 10 RSHIFT F AND OR
        SWAP FFF000F0 AND OR
        00000090 OR hw ! #,
        ;
: ,MLA  hw @ DUP 10 * FF000 AND
        OVER 100 * F00 AND OR
        OVER 10 RSHIFT F AND OR
        SWAP FFF000F0 AND OR
        00200090 OR hw ! #,
        ;

00800090 simple_inst ,UMULL
00A00090 simple_inst ,UMLAL
00C00090 simple_inst ,SMULL
00E00090 simple_inst ,SMLAL

( Registers of ARM7 eForth VM )

0 register IP
1 register SP
2 register RP
3 register UP
4 register WP
5 register TOS
6 register XP
7 register YP
8 register WR
9 register HR
A register XD
B register YD
C register ER
D register CTR
E register LP
F register PC

0 register IP,
1 register SP,
2 register RP,
3 register UP,
4 register WP,
5 register TOS,
6 register XP,
7 register YP,
8 register WR,
9 register HR,
A register XD,
B register YD,
C register ER,
D register CTR,
E register LP,
F register PC,

( eForth model )

: (makehead)
   20 WORD                                \ get name of new definition
   lasth @ ,                         \ fill link field of last word
   h @ lasth !                        \ save nfa in lasth
   COUNT DUP ,b                                \ store count
   FOR AFT
      COUNT ,b                              \ fill name field
   THEN NEXT
   DROP ,b0 
   h @ lastc !
   ;

: makehead
   >IN @ >R                               \ save interpreter pointer
   (makehead)
   R> >IN !                               \ restore word pointer
   ;

( structured assembler )

: $LIT ( -- )
   22 WORD
   COUNT DUP ,b ( compile count )
   FOR AFT
      COUNT ,b ( compile characters )
   THEN NEXT
   DROP ,b0 ;

: bra     AL, ,B ;
: bz      EQ, ,B ;
: bc      CS, ,B ;
: bnz     NE, ,B ;
: bnc     CC, ,B ;

: begin   h @ ;
: until   bz ;
: untilz  bnz ;
: untilnc bc ;
: jmp     bra ;

: if      h @ 0 bz ;
: ifz     h @ 0 bnz ;
: ifc     h @ 0 bnc ;
: ifnc    h @ 0 bc ;
: skip    h @ 0 bra ;
: then    >R begin R@ - 4 / 2 - $FFFFFF AND
	  R@ ram@ $FF000000 AND OR R> ram! ;
: else    skip  SWAP then ;
: while   if SWAP ;
: whilenc ifnc  SWAP ;
: repeat  jmp then ;
: again   jmp ;
: aft ( a -- a' a" )
          DROP skip begin SWAP ;

: ':   begin  .head CONSTANT [FORTH] IMMEDIATE [META] DOES> anew R> @  ,BL ;
: CODE  makehead ': anew ;            \ for eforth kernel words
: code  makehead ': anew ;            \ for eforth kernel words

: END-CODE ;
